"""
This module is modified from the original code of the paper:

- Gargiulo, et.al. Task Singular Vectors: Reducing Task Interference in Model Merging
    - http://arxiv.org/abs/2412.00081
    - https://github.com/AntoAndGar/task_singular_vectors/
"""

from .TSVM import TaskSingularVectorMerging
