from .imagenet import templates  # noqa: F401

classnames_full = [
    "goldfish, Carassius auratus",
    "European fire salamander, Salamandra salamandra",
    "bullfrog, Rana catesbeiana",
    "tailed frog, bell toad, ribbed toad, tailed toad, Ascaphus trui",
    "American alligator, Alligator mississipiensis",
    "boa constrictor, Constrictor constrictor",
    "trilobite",
    "scorpion",
    "black widow, Latrodectus mactans",
    "tarantula",
    "centipede",
    "koala, koala bear, kangaroo bear, native bear, Phascolarctos cinereus",
    "jellyfish",
    "brain coral",
    "snail",
    "sea slug, nudibranch",
    "American lobster, Northern lobster, Maine lobster, Homarus americanus",
    "spiny lobster, langouste, rock lobster, crawfish, crayfish, sea crawfish",
    "black stork, Ciconia nigra",
    "king penguin, Aptenodytes patagonica",
    "albatross, mollymawk",
    "dugong, Dugong dugon",
    "Yorkshire terrier",
    "golden retriever",
    "Labrador retriever",
    "German shepherd, German shepherd dog, German police dog, alsatian",
    "standard poodle",
    "tabby, tabby cat",
    "Persian cat",
    "Egyptian cat",
    "cougar, puma, catamount, mountain lion, painter, panther, Felis concolor",
    "lion, king of beasts, Panthera leo",
    "brown bear, bruin, Ursus arctos",
    "ladybug, ladybeetle, lady beetle, ladybird, ladybird beetle",
    "grasshopper, hopper",
    "walking stick, walkingstick, stick insect",
    "cockroach, roach",
    "mantis, mantid",
    "dragonfly, darning needle, devil's darning needle, sewing needle, snake feeder, snake doctor, mosquito hawk, skeeter hawk",
    "monarch, monarch butterfly, milkweed butterfly, Danaus plexippus",
    "sulphur butterfly, sulfur butterfly",
    "sea cucumber, holothurian",
    "guinea pig, Cavia cobaya",
    "hog, pig, grunter, squealer, Sus scrofa",
    "ox",
    "bison",
    "bighorn, bighorn sheep, cimarron, Rocky Mountain bighorn, Rocky Mountain sheep, Ovis canadensis",
    "gazelle",
    "Arabian camel, dromedary, Camelus dromedarius",
    "orangutan, orang, orangutang, Pongo pygmaeus",
    "chimpanzee, chimp, Pan troglodytes",
    "baboon",
    "African elephant, Loxodonta africana",
    "lesser panda, red panda, panda, bear cat, cat bear, Ailurus fulgens",
    "abacus",
    "academic gown, academic robe, judge's robe",
    "altar",
    "backpack, back pack, knapsack, packsack, rucksack, haversack",
    "bannister, banister, balustrade, balusters, handrail",
    "barbershop",
    "barn",
    "barrel, cask",
    "basketball",
    "bathtub, bathing tub, bath, tub",
    "beach wagon, station wagon, wagon, estate car, beach waggon, station waggon, waggon",
    "beacon, lighthouse, beacon light, pharos",
    "beaker",
    "beer bottle",
    "bikini, two-piece",
    "binoculars, field glasses, opera glasses",
    "birdhouse",
    "bow tie, bow-tie, bowtie",
    "brass, memorial tablet, plaque",
    "bucket, pail",
    "bullet train, bullet",
    "butcher shop, meat market",
    "candle, taper, wax light",
    "cannon",
    "cardigan",
    "cash machine, cash dispenser, automated teller machine, automatic teller machine, automated teller, automatic teller, ATM",
    "CD player",
    "chest",
    "Christmas stocking",
    "cliff dwelling",
    "computer keyboard, keypad",
    "confectionery, confectionary, candy store",
    "convertible",
    "crane",
    "dam, dike, dyke",
    "desk",
    "dining table, board",
    "dumbbell",
    "flagpole, flagstaff",
    "fly",
    "fountain",
    "freight car",
    "frying pan, frypan, skillet",
    "fur coat",
    "gasmask, respirator, gas helmet",
    "go-kart",
    "gondola",
    "hourglass",
    "iPod",
    "jinrikisha, ricksha, rickshaw",
    "kimono",
    "lampshade, lamp shade",
    "lawn mower, mower",
    "lifeboat",
    "limousine, limo",
    "magnetic compass",
    "maypole",
    "military uniform",
    "miniskirt, mini",
    "moving van",
    "neck brace",
    "obelisk",
    "oboe, hautboy, hautbois",
    "organ, pipe organ",
    "parking meter",
    "pay-phone, pay-station",
    "picket fence, paling",
    "pill bottle",
    "plunger, plumber's helper",
    "police van, police wagon, paddy wagon, patrol wagon, wagon, black Maria",
    "poncho",
    "pop bottle, soda bottle",
    "potter's wheel",
    "projectile, missile",
    "punching bag, punch bag, punching ball, punchball",
    "refrigerator, icebox",
    "remote control, remote",
    "rocking chair, rocker",
    "rugby ball",
    "sandal",
    "school bus",
    "scoreboard",
    "sewing machine",
    "snorkel",
    "sock",
    "sombrero",
    "space heater",
    "spider web, spider's web",
    "sports car, sport car",
    "steel arch bridge",
    "stopwatch, stop watch",
    "sunglasses, dark glasses, shades",
    "suspension bridge",
    "swimming trunks, bathing trunks",
    "syringe",
    "teapot",
    "teddy, teddy bear",
    "thatch, thatched roof",
    "torch",
    "tractor",
    "triumphal arch",
    "trolleybus, trolley coach, trackless trolley",
    "turnstile",
    "umbrella",
    "vestment",
    "viaduct",
    "volleyball",
    "water jug",
    "water tower",
    "wok",
    "wooden spoon",
    "comic book",
    "reel",
    "guacamole",
    "ice cream, icecream",
    "ice lolly, lolly, lollipop, popsicle",
    "goose",
    "drumstick",
    "plate",
    "pretzel",
    "mashed potato",
    "cauliflower",
    "bell pepper",
    "lemon",
    "banana",
    "pomegranate",
    "meat loaf, meatloaf",
    "pizza, pizza pie",
    "potpie",
    "espresso",
    "bee",
    "apron",
    "pole",
    "Chihuahua",
    "alp",
    "cliff, drop, drop-off",
    "coral reef",
    "lakeside, lakeshore",
    "seashore, coast, seacoast, sea-coast",
    "acorn",
    "broom",
    "mushroom",
    "nail",
    "chain",
    "slug",
    "orange",
]

classnames = [n.split(",")[0] for n in classnames_full]

assert len(classnames) == 200, "Tiny ImageNet classnames should have 200 classes"
