classnames = [
    "abyssinian",
    "american_bulldog",
    "american_pit_bull_terrier",
    "basset_hound",
    "beagle",
    "bengal",
    "birman",
    "bombay",
    "boxer",
    "british_shorthair",
    "chihuahua",
    "egyptian_mau",
    "english_cocker_spaniel",
    "english_setter",
    "german_shorthaired",
    "great_pyrenees",
    "havanese",
    "japanese_chin",
    "keeshond",
    "leonberger",
    "maine_coon",
    "miniature_pinscher",
    "newfoundland",
    "persian",
    "pomeranian",
    "pug",
    "ragdoll",
    "russian_blue",
    "saint_bernard",
    "samoyed",
    "scottish_terrier",
    "shiba_inu",
    "siamese",
    "sphynx",
    "staffordshire_bull_terrier",
    "wheaten_terrier",
    "yorkshire_terrier",
]

templates = [lambda c: f"a photo of a {c}, a type of pet."]
